/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package com.ebmwebsourcing.jbi.adaptor.impl;

import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.CDK_NAMESPACE_PREFIX;
import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.CDK_NAMESPACE_URI;
import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.JBI_NAMESPACE_PREFIX;
import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.JBI_NAMESPACE_URI;
import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.XSI_NAMESPACE_PREFIX;
import static com.ebmwebsourcing.jbi.adaptor.impl.JbiConstants.XSI_NAMESPACE_URI;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.ow2.petals.component.api.ServiceConfiguration;
import org.ow2.petals.component.api.ServiceConfiguration.ServiceType;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.Transformers;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class ServiceUnitJbiGenerator {

    private static final String SERVICE_NAMESPACE_PREFIX = "srvNs";

    private static final String INTERFACE_NAMESPACE_PREFIX = "itfNs";

    private final ServiceConfiguration serviceConfiguration;

    private final File installationRoot;

    public ServiceUnitJbiGenerator(ServiceConfiguration serviceConfiguration, File installationRoot) {
        assert serviceConfiguration != null;
        assert installationRoot != null;
        this.serviceConfiguration = serviceConfiguration;
        this.installationRoot = installationRoot;
    }

    /**
     * Generate the jbi.xml file.
     * 
     * @param multiservice
     *            If true, services dependencies of the
     *            {@link ServiceConfiguration} will be generate in the same
     *            jbi.xml file.
     * @return
     */
    public final File generate(boolean multiservice) {
        File serviceUnitMetaInfDir = new File(installationRoot, "META-INF");
        serviceUnitMetaInfDir.mkdirs();
        File jbiDumpTarget = new File(serviceUnitMetaInfDir, "jbi.xml");
        final Result result = new StreamResult(jbiDumpTarget);

        Document document = DocumentBuilders.newDocument();
        document.setXmlVersion("1.0");
        document.setXmlStandalone(true);

        Map<String, String> jbiEltAttr = new HashMap<String, String>();
        jbiEltAttr.put("version", "1.0");
        Element jbiElement = addElement(document, document, JBI_NAMESPACE_URI, JBI_NAMESPACE_PREFIX
                + ":jbi", jbiEltAttr);
        jbiElement.setAttribute(XMLConstants.XMLNS_ATTRIBUTE + ":" + XSI_NAMESPACE_PREFIX,
                XSI_NAMESPACE_URI);

        Map<String, String> servicesEltAttr = new HashMap<String, String>();
        servicesEltAttr.put("binding-component", "true");
        
        Element servicesElement = addElement(document, jbiElement, JBI_NAMESPACE_URI,
                "jbi:services", servicesEltAttr);

        generateServiceTypeElement(document, servicesElement, serviceConfiguration);
        if(multiservice) {
            for(ServiceConfiguration s : this.serviceConfiguration.getServiceConfigurationDependencies()) {
                generateServiceTypeElement(document, servicesElement, s);
            }
        }
        Source domSource = new DOMSource(document);
        Transformer transformer = Transformers.takeTransformer();
        try {
            transformer.transform(domSource, result);
            return jbiDumpTarget;
        } catch (TransformerException e) {
            throw new UncheckedException("Impossible to generate the JBI file", e);
        } finally {
            Transformers.releaseTransformer(transformer);
        }
    }

    /**
     * Generate provide or consume element
     * @param document
     * @param servicesElement
     * @param serviceConfiguration
     */
    private void generateServiceTypeElement(Document document, Element servicesElement,
            ServiceConfiguration serviceConfiguration) {
        String serviceTypeElementName = "jbi:provides";
        if (ServiceType.CONSUME.equals(serviceConfiguration.getServiceType())) {
            serviceTypeElementName = "jbi:consumes";
        }
        // Create attributes
        Map<String, String> serviceTypeEltAttr = new HashMap<String, String>();
        serviceTypeEltAttr.put("endpoint-name", serviceConfiguration.getEndpointName());
        serviceTypeEltAttr.put("interface-name", INTERFACE_NAMESPACE_PREFIX + ":"
                + serviceConfiguration.getInterfaceQName().getLocalPart());
        serviceTypeEltAttr.put("service-name", SERVICE_NAMESPACE_PREFIX + ":"
                + serviceConfiguration.getServiceQName().getLocalPart());
        serviceTypeEltAttr.put(XMLConstants.XMLNS_ATTRIBUTE + ":" + SERVICE_NAMESPACE_PREFIX,
                serviceConfiguration.getServiceQName().getNamespaceURI());
        serviceTypeEltAttr.put(XMLConstants.XMLNS_ATTRIBUTE + ":" + INTERFACE_NAMESPACE_PREFIX,
                serviceConfiguration.getInterfaceQName().getNamespaceURI());
        Element providesElement = addElement(document, servicesElement, JBI_NAMESPACE_URI,
                serviceTypeElementName, serviceTypeEltAttr);
        
        if (ServiceType.PROVIDE.equals(serviceConfiguration.getServiceType())) {
            if (serviceConfiguration.getServiceDescription() != null) {
                addElement(document, providesElement, CDK_NAMESPACE_URI, CDK_NAMESPACE_PREFIX
                        + ":wsdl", serviceConfiguration.getServiceDescription().toString());
            } else {
                addElement(document, providesElement, CDK_NAMESPACE_URI, CDK_NAMESPACE_PREFIX
                        + ":wsdl");
            }
        } else {
            // If no mep is defined, add a default mep
            if (serviceConfiguration.getParameter("{" + CDK_NAMESPACE_URI + "}mep") == null) {
                Map<String, String> mepAttr = new HashMap<String, String>();
                mepAttr.put(XSI_NAMESPACE_PREFIX + ":nil", "true");
                addElement(document, providesElement, CDK_NAMESPACE_URI, CDK_NAMESPACE_PREFIX
                        + ":mep", mepAttr);
            }
        }
        //append service parameters
        for (String specificParameterName : serviceConfiguration.getParameterNames()) {
            if (specificParameterName.equals("{" + CDK_NAMESPACE_URI + "}operation")) {
                String operationQName = serviceConfiguration.getParameter(specificParameterName);
                addElement(
                        document,
                        providesElement,
                        CDK_NAMESPACE_URI,
                        CDK_NAMESPACE_PREFIX + ":operation",
                        SERVICE_NAMESPACE_PREFIX + ":"
                                + operationQName.substring(operationQName.lastIndexOf("}") + 1));
            } else {
                QName specificParameterQName = QName.valueOf(specificParameterName);
                addElement(document, providesElement, specificParameterQName.getNamespaceURI(),
                        specificParameterQName.getLocalPart(),
                        serviceConfiguration.getParameter(specificParameterName));
            }
        }
    }

    private static final Element addElement(Document document, Node parentNode,
            String namespaceUri, String qualifiedName) {
        return addElement(document, parentNode, namespaceUri, qualifiedName, null,
                new HashMap<String, String>());
    }

    private static final Element addElement(Document document, Node parentNode,
            String namespaceUri, String qualifiedName, Map<String, String> attributes) {
        return addElement(document, parentNode, namespaceUri, qualifiedName, null, attributes);
    }

    private static final Element addElement(Document document, Node parentNode,
            String namespaceUri, String qualifiedName, String value) {
        return addElement(document, parentNode, namespaceUri, qualifiedName, value,
                new HashMap<String, String>());
    }

    private static final Element addElement(Document document, Node parentNode,
            String namespaceUri, String qualifiedName, String value, Map<String, String> attributes) {
        Element newElement = document.createElementNS(namespaceUri, qualifiedName);
        if (value != null) {
            newElement.setTextContent(value);
        }
        for (String attributeName : attributes.keySet()) {
            newElement.setAttribute(attributeName, attributes.get(attributeName));
        }
        parentNode.appendChild(newElement);
        return newElement;
    }

}
