/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package com.ebmwebsourcing.jbi.adaptor.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.jar.JarFile;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.apache.commons.io.FileUtils;
import org.ow2.petals.component.api.Component;
import org.ow2.petals.component.api.ComponentConfiguration;
import org.ow2.petals.component.api.Message;
import org.ow2.petals.component.api.ServiceConfiguration;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.JBIBootstrap;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.JBIProcessorManagerObserver;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;

import com.ebmwebsourcing.easycommons.io.FileSystemHelper;
import com.ebmwebsourcing.easycommons.io.IOHelper;
import com.ebmwebsourcing.easycommons.io.ZipHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.jbi.adaptor.impl.mock.MockComponentContext;
import com.ebmwebsourcing.jbi.adaptor.impl.mock.MockDeliveryChannel;
import com.ebmwebsourcing.jbi.adaptor.impl.mock.MockInstallationContext;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class WrappedComponent implements Component {

	private static final List<String> startedComponents = new ArrayList<String>();
	
    private static final String JBI_DIR_NAME = "jbi";

    private final File componentRepositoryRoot;

    private final ComponentType componentType;

    private final ComponentConfiguration componentConfiguration;

    private final Jbi jbi;

    private final AbstractComponent component;

    private MockComponentContext mockComponentContext;

    public WrappedComponent(ComponentType componentType,
            ComponentConfiguration componentConfiguration) {
        assert componentConfiguration != null;
        this.componentRepositoryRoot = createComponentRepositoryRoot();
        System.out.println(String.format("------------- Created component repository root '%s'",
                componentRepositoryRoot));
        this.componentType = componentType;
        this.componentConfiguration = componentConfiguration;
        this.jbi = generateComponentJbi();
        copyOthersJbiResources();
        copyJbiResources(componentConfiguration.getResources(), getInstallRoot());
        this.component = createComponent(jbi.getComponent().getComponentClassName().getContent());
        initialize();
    }

    private final File getInstallRoot() {
        return new File(componentRepositoryRoot, "install");
    }

    private final File getWorkspaceRoot() {
        return new File(componentRepositoryRoot, "work");
    }

    private Jbi generateComponentJbi() {
        ComponentJbiGenerator componentJbiGenerator = new ComponentJbiGenerator(componentType,
                componentConfiguration, getInstallRoot());

        componentJbiGenerator.generate();
        File generatedJbiFile = new File(getInstallRoot() + File.separator + "META-INF"
                + File.separator, "jbi.xml");
        assert generatedJbiFile.exists();
        FileInputStream fis;
        try {
            fis = new FileInputStream(generatedJbiFile);
        } catch (FileNotFoundException fnfe) {
            throw new UncheckedException(fnfe);
        }
        assert generatedJbiFile.exists();
        try {
            return CDKJBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(fis);
        } catch (JBIDescriptorException cjde) {
            throw new UncheckedException(cjde);
        } finally {
            IOHelper.close(fis);
        }
    }

    private final String getComponentName() {
        return jbi.getComponent().getIdentification().getName();
    }

    private void copyOthersJbiResources() {
        // TODO comment on this!

        // TODO :
        // 1) convert class name to path like
        // org/ow2/petals/bc/filetransfer/FileTrasnferComponent

        // 2) getResource and keep obatained jar
        // 3) keep only jbi resource below that starts with jar name foudn in 2

        String componentClassName = componentType.getClassName();
        String componentClassFilePath = componentClassName.replaceAll("\\.", "/") + ".class";
        URL componentClassFileUrl = getClass().getClassLoader().getResource(componentClassFilePath);
        assert componentClassFileUrl != null;
        String componentClassFileUrlStr = componentClassFileUrl.toString();
        String componentBasePathPrefix = componentClassFileUrlStr.substring(0,
                componentClassFileUrlStr.indexOf(componentClassFilePath));

        URL executionDirUrl = null;
        try {
            Enumeration<URL> jbiFolderUrls = getClass().getClassLoader().getResources(JBI_DIR_NAME);
            while (jbiFolderUrls.hasMoreElements()) {
                URL jbiFolderUrl = jbiFolderUrls.nextElement();
                if (jbiFolderUrl.toString().startsWith(componentBasePathPrefix)) {
                    executionDirUrl = jbiFolderUrl;
                }
            }
        } catch (IOException ioe) {
            throw new UncheckedException(ioe);
        }
        assert executionDirUrl != null;

        try {
            if (executionDirUrl.getProtocol().equals("jar")) {

                String executionDirStr = executionDirUrl.toString();
                int executionJarEndIndex = executionDirStr.indexOf(".jar!") + ".jar".length();
                assert executionJarEndIndex != -1;

                String executionJarStr = executionDirStr.substring("jar:file:".length(),
                        executionJarEndIndex);
                JarFile executionJar = new JarFile(executionJarStr);

                File jbiDirExtractedFromComponentJar = FileSystemHelper
                        .createTempDir("jbiDirExtractedFromComponentJar");
                ZipHelper.extractEntriesFromArchive(executionJar, JBI_DIR_NAME + File.separator,
                        jbiDirExtractedFromComponentJar);
                copyAllJbiDirContentExceptJbiXmlFile(new File(jbiDirExtractedFromComponentJar,
                        JBI_DIR_NAME));
            } else {
                File executionDir = new File(executionDirUrl.toURI());
                copyAllJbiDirContentExceptJbiXmlFile(executionDir);
            }
        } catch (Exception e) {
            throw new UncheckedException(e);
        }
    }

    private void copyAllJbiDirContentExceptJbiXmlFile(File jbiDir) throws IOException {
        for (File fileInJbiDir : jbiDir.listFiles()) {
            if ("jbi.xml".equals(fileInJbiDir.getName())) {
                // jbi.xml was generated just before, skip it.
                continue;
            }
            if (fileInJbiDir.isDirectory()) {
                FileUtils.copyDirectoryToDirectory(fileInJbiDir, getInstallRoot());
            } else {
                FileUtils.copyFileToDirectory(fileInJbiDir, getInstallRoot());
            }
        }
    }

    private void copyJbiResources(URL[] resourceUrls, File installationRoot) {
        for (URL resourceUrl : resourceUrls) {
            try {
                if (resourceUrl.getProtocol().equals("jar")) {
                    // TODO : not supported, to be implemented... otherwise
                    // resource copying when resource
                    // in jar will not work
                    continue;
                }
                File resourceFile = new File(resourceUrl.toURI());
                if (resourceFile.isDirectory()) {
                    FileUtils.copyDirectory(resourceFile, installationRoot);
                } else {
                    FileUtils.copyFileToDirectory(resourceFile, installationRoot);
                }
            } catch (Exception e) {
                throw new UncheckedException(e);
            }
        }
    }

    private AbstractComponent createComponent(String componentClassName) {
        try {
            return (AbstractComponent) Class.forName(componentClassName).newInstance();
        } catch (Exception e) {
            throw new UncheckedException(e);
        }
    }

    private static final File createComponentRepositoryRoot() {
        try {
            return FileSystemHelper.createTempDir("repository");
        } catch (IOException ioe) {
            throw new UncheckedException(ioe);
        }
    }

    @SuppressWarnings("unchecked")
    public void initialize() {
        this.mockComponentContext = new MockComponentContext(jbi.getComponent().getIdentification()
                .getName(), this.component, getInstallRoot().getAbsolutePath(), getWorkspaceRoot()
                .getAbsolutePath(), componentConfiguration.getLogger());
        InstallationContext mockInstallationContext = new MockInstallationContext(getInstallRoot(),
                jbi, mockComponentContext, true);
        // jbi bootstrap
        try {
            Class<? extends JBIBootstrap> jbiBootstrapClass = (Class<? extends JBIBootstrap>) Class
                    .forName(jbi.getComponent().getBootstrapClassName());
            JBIBootstrap jbiBootstrap = jbiBootstrapClass.newInstance();
            jbiBootstrap.init(mockInstallationContext);

        } catch (Exception e) {
            throw new UncheckedException(e);
        }

        // jbi initialize
        try {
            this.component.init(mockInstallationContext.getContext());
        } catch (JBIException jbie) {
            throw new UncheckedException(jbie);
        }

    }

    public void start() {
        try {
            this.component.start();
            startedComponents.add(getComponentName());
        } catch (JBIException e) {
            throw new UncheckedException("Impossible to start the wrapped component", e);
        }
    }


    public boolean isStarted() {
        return startedComponents.contains(getComponentName());
    }

    public void stop() {
        try {
            this.component.stop();

            final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
            if (serviceUnitManager != null) {
                uninstallAllServices();
            }

            this.component.shutDown();
            startedComponents.remove(getComponentName());
        } catch (JBIException e) {
            throw new UncheckedException("Impossible to stop the wrapped component", e);
        }
    }

    public boolean isStopped() {
        return !isStarted();
    }

    public void installService(ServiceConfiguration serviceConfiguration) {
        try {
            AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
            assert serviceUnitManager != null;
            // TODO : create SA/install in repository - ranger les repos...
            File serviceAssembliesInstallRoot = new File(componentRepositoryRoot,
                    "service-assembliesInstall");

            String suName = serviceConfiguration.getConfigurationName();
            File serviceUnitDir = new File(serviceAssembliesInstallRoot, suName);

            copyJbiResources(serviceConfiguration.getResources(), serviceUnitDir);

            ServiceUnitJbiGenerator serviceUnitJbiGenerator = new ServiceUnitJbiGenerator(
                    serviceConfiguration, serviceUnitDir);
            boolean multiservice = false;
            if(ComponentType.SE_EIP.equals(componentType)) {
                multiservice=true;
            }
            serviceUnitJbiGenerator.generate(multiservice);

            serviceUnitManager.deploy(suName, serviceUnitDir.getAbsolutePath());
            serviceUnitManager.init(suName, serviceUnitDir.getAbsolutePath());
            serviceUnitManager.start(suName);
        } catch (DeploymentException e) {
            throw new UncheckedException("Impossible to install the specified service", e);
        }
    }
    
    public void uninstallService(ServiceConfiguration serviceConfiguration) {
		try {
            AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();

            assert serviceUnitManager != null;

			String suName = serviceConfiguration.getConfigurationName();
            ServiceUnitDataHandler serviceUnitDataHandler = serviceUnitManager.getSUDataHandler(suName);
			serviceUnitManager.stop(suName);
			serviceUnitManager.shutDown(suName);
			serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
		} catch (DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall the specified service", e);
		}
    }

	@Override
	public void uninstallAllServices() {
		try {
            AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
	        
            assert serviceUnitManager != null;

            for (ServiceUnitDataHandler serviceUnitDataHandler : serviceUnitManager.getServiceUnitDataHandlers()) {
				String suName = serviceUnitDataHandler.getName();
				serviceUnitManager.stop(suName);
				serviceUnitManager.shutDown(suName);
				serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
	        }
		} catch (DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall all the services", e);
		}
	}
	
    public boolean isServiceInstalled(ServiceConfiguration serviceConfiguration) {
        AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();

        assert serviceUnitManager != null;

        String suName = serviceConfiguration.getConfigurationName();
        return serviceUnitManager.getSUDataHandler(suName) != null;
    }

    public void processMessageFromServiceBus(Message message) {
        assert message instanceof AbstractWrappedToComponentMessage;
        AbstractWrappedToComponentMessage wrappedMessage = (AbstractWrappedToComponentMessage) message;
        try {
            final MessageExchange msgExchange = wrappedMessage.getMessageExchange();

            final Semaphore sem = new Semaphore(0);
            final JBIProcessorManager processorManager = this.component.getProcessorManager();
            JBIProcessorManagerObserver jbiProccesorManagerObserver = new JBIProcessorManagerObserver() {
                @Override
                public void beforeProcessingMessage(final Exchange exchange) {
                    // do nothing
                }

                @Override
                public void afterProcessingMessage(final Exchange exchange) {
                    // this will unblock the execution of pushSync until
                    // the message has been processed by the provider
                    if (exchange.getExchangeId().equals(msgExchange.getExchangeId())) {
                        sem.release();
                    }
                }
            };
            processorManager.addObserver(jbiProccesorManagerObserver);

            ((MockDeliveryChannel) this.component.getContext().getDeliveryChannel())
                    .enqueueIncomingMessage(msgExchange);

            sem.acquire();

            processorManager.deleteObserver(jbiProccesorManagerObserver);
        } catch (InterruptedException e) {
            throw new UncheckedException("Impossible to send a message from kernel", e);
        } catch (MessagingException e) {
            throw new UncheckedException("Impossible to send a message from kernel", e);
        }
    }

    private final void pushMessageInDeliveryChannel(Message message) {
        AbstractWrappedToComponentMessage wrappedMessage = (AbstractWrappedToComponentMessage) message;

        try {
            MessageExchange msgExchange = wrappedMessage.getMessageExchange();

            ((MockDeliveryChannel) this.component.getContext().getDeliveryChannel()).enqueueIncomingMessage(msgExchange);
        } catch (MessagingException e) {
            throw new UncheckedException("Impossible to push a message", e);
        }
    }
    
    @Override
    public void pushRequestToProvider(Message message) {
    	pushMessageInDeliveryChannel(message);
    }
    
    @Override
    public void pushResponseToConsumer(Message message) {
    	pushMessageInDeliveryChannel(message);
    }

    @Override
    public Message pollResponseFromProvider() {
        try {
            MessageExchange msgExchange = ((MockDeliveryChannel) this.component.getContext()
                    .getDeliveryChannel()).dequeueResponseToClient();
            assert !MEPConstants.IN_ONLY_PATTERN.equals(msgExchange.getPattern()) :
            	"Polling response from provider while processing an in only exchange is meaningless.";

            return new WrappedResponseFromProviderMessage(msgExchange);
        } catch (MessagingException e) {
            throw new UncheckedException("Impossible to poll response to consumer message", e);
        }
    }

    @Override
    public Message pollRequestFromConsumer() {
        try {
            MessageExchange msgExchange = ((MockDeliveryChannel) this.component.getContext()
                    .getDeliveryChannel()).dequeueRequestToServer();

            return new WrappedRequestFromConsumerMessage(msgExchange);
        } catch (MessagingException e) {
            throw new UncheckedException("Impossible to poll request to provider message", e);
        }
    }
}
