/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package com.ebmwebsourcing.jbi.adaptor.impl.mock;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.mockito.Mockito;
import org.ow2.petals.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.impl.DefaultMessageExchangeFactory;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * @author Adrien Ruffie - EBM WebSourcing
 */
public class MockDeliveryChannel implements DeliveryChannel {

    private final Logger logger;

    private final LinkedBlockingQueue<MessageExchange> incomingMessagesQueue;

    private final LinkedBlockingQueue<MessageExchange> responsesToClientQueue;

    private final LinkedBlockingQueue<MessageExchange> requestsToServerQueue;

    private String synchronousExchangeId;

    private final Semaphore synchronousExchangeSemaphore = new Semaphore(0);

    private final PetalsDeliveryChannel mockPetalsChannel;

    public MockDeliveryChannel(final Logger logger) {
        this.logger = logger;
        this.incomingMessagesQueue = new LinkedBlockingQueue<MessageExchange>();
        this.responsesToClientQueue = new LinkedBlockingQueue<MessageExchange>();
        this.requestsToServerQueue = new LinkedBlockingQueue<MessageExchange>();
        // we have to create this adapter, because jbi-adapter-impl is really old and not aligned (like
        // petals-cdk-junit) with the PetalsMessageExchange and stuffs, so it is too much work to convert this class to
        // PetalsDeliveryChannel
        this.mockPetalsChannel = Mockito.mock(PetalsDeliveryChannel.class);
        final ComponentContext ccMock = Mockito.mock(ComponentContext.class);
        final Location location = new Location("test", "test");
        Mockito.when(this.mockPetalsChannel.getLogger()).thenReturn(this.logger);
        Mockito.when(this.mockPetalsChannel.getContext()).thenReturn(ccMock);
        Mockito.when(this.mockPetalsChannel.getLocation()).thenReturn(location);
        Mockito.when(this.mockPetalsChannel.isOpened()).thenReturn(true);
    }

    public void enqueueIncomingMessage(final MessageExchange msgExchange) {
        try {
            this.incomingMessagesQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public void enqueueRequestToServer(final MessageExchange msgExchange) {
        try {
            this.requestsToServerQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public void enqueueResponseToClient(final MessageExchange msgExchange) {
        try {
            this.responsesToClientQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueIncomingMessage() {
        try {
            return this.incomingMessagesQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueRequestToServer() {
        try {
            return this.requestsToServerQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueResponseToClient() {
        try {
            return this.responsesToClientQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    @Override
    public MessageExchange accept() throws MessagingException {
        try {
            MessageExchange exchange = this.incomingMessagesQueue.take();
            if (exchange.getExchangeId().equals(synchronousExchangeId)) {
                synchronousExchangeSemaphore.release();
                return accept();
            } else {
                return exchange;
            }
        } catch (final InterruptedException e) {
            throw new MessagingException(e);
        }
    }

    @Override
    public MessageExchange accept(final long timeoutMS) throws MessagingException,
            IllegalArgumentException {
        return null;
    }

    @Override
    public void close() throws MessagingException {
        // TODO Auto-generated method stub
    }

    @Override
    public MessageExchangeFactory createExchangeFactory() {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory.create(this.mockPetalsChannel);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory
                .createForInterface(this.mockPetalsChannel, interfaceName);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory
                .createForEndpoint(this.mockPetalsChannel, endpoint);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory
                .createForService(this.mockPetalsChannel, serviceName);
        return exchangeFactoryImpl;
    }

    @Override
    public void send(final MessageExchange messageExchange) throws MessagingException {
        if (messageExchange.getEndpoint() == null) {
            resolveEndpoint(messageExchange);
        }

        final MessageExchangeWrapper msgExchange = (MessageExchangeWrapper) messageExchange;

        try {
            if (messageExchange.getRole().equals(Role.PROVIDER)) {
                this.responsesToClientQueue.put(msgExchange.getMessageExchange());
            } else {
                this.requestsToServerQueue.put(msgExchange.getMessageExchange());
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    private void resolveEndpoint(MessageExchange messageExchange) throws MessagingException {
        ServiceEndpoint endpoint = MockComponentContext.resolveServiceEndpoint(messageExchange
                .getService());
        if (endpoint != null) {
            messageExchange.setEndpoint(endpoint);
        } else {
            throw new MessagingException("Endpoint not avalaible");
        }
    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange) throws MessagingException {
        return this.sendSync(msgExchange, 0);
    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange, final long timeoutMS)
            throws MessagingException {
        boolean returnValue = false;
        synchronousExchangeId = msgExchange.getExchangeId();
        this.send(msgExchange);
        try {
            if (timeoutMS > 0) {
                returnValue = synchronousExchangeSemaphore.tryAcquire(timeoutMS,
                        TimeUnit.MILLISECONDS);

            } else {
                synchronousExchangeSemaphore.acquire();
                returnValue = true;
            }
        } catch (InterruptedException e) {
            throw new UncheckedException(e);
        }
        return returnValue;
    }

}
